from rkale.utils import check_files, check_paths, read, sync


def get_input(verbose_info):
    answer = input(
        "Are you sure you want to continue? [ls/y/N] (ls to expand changes) "
    )
    if answer == "ls":
        print(verbose_info)
        answer = input("Are you sure you want to continue? [y/N] ")
    return answer in ["y", "yes"]


def get_answers(operation, checks, paths, force=False):
    answers = []
    for check, (source, destination) in zip(checks, paths):
        missing_src, missing_dst = read(check.src_out), read(check.dst_out)
        if check.stderr:
            print(check.stderr)
            answers.append(False)
        elif force:
            answers.append(True)
        else:
            verbose_info = []
            print(f"\nChanges to be applied to {destination}:")
            if missing_dst:
                print(f"{len(missing_dst)} new files will be added")
                verbose_info.extend(["Files to be added:\n", *missing_dst])
            if operation == "sync" and missing_src:
                print(f"WARNING! {len(missing_src)} files will be removed")
                verbose_info.extend(["Files to be removed:\n", *missing_src])

            if (operation == "sync" and missing_src) or missing_dst:
                answers.append(get_input("".join(verbose_info)))
            else:
                print(f"Source and destination match, no files to {operation}")
                answers.append(False)
    return answers


def handle_copy(paths, force=False):
    check_paths(paths)
    with check_files(paths) as checks:
        answers = get_answers("copy", checks, paths, force=force)
        for answer, check, (source, destination) in zip(answers, checks, paths):
            if answer:
                print(f"Copying files to {destination}...")
                sync(source, destination, files_from=check.dst_out, progress=True)


def handle_sync(paths, force=False):
    check_paths(paths)
    with check_files(paths) as checks:
        answers = get_answers("sync", checks, paths, force=force)
        for answer, check, (source, destination) in zip(answers, checks, paths):
            if answer:
                if read(check.src_out):
                    print(f"Removing files from {destination}...")
                    sync(source, destination, files_from=check.src_out, progress=False)
                if read(check.dst_out):
                    print(f"Copying files to {destination}...")
                    sync(source, destination, files_from=check.dst_out, progress=True)



import os
import re
import subprocess
import tempfile
from collections import namedtuple
from contextlib import contextmanager, suppress

from tqdm import tqdm


def read(file_path):
    if os.path.exists(file_path):
        return open(file_path, "r").readlines()
    return []


@contextmanager
def check_files(paths):
    Check = namedtuple("Check", ["src_out", "dst_out", "stderr"])
    checks = []
    for source, destination in paths:
        print(f"Comparing paths {source} and {destination}...")
        src_out, dst_out, error_out = (tempfile.mkstemp()[1] for _ in range(3))
        result = subprocess.run(
            f"rclone check {source} {destination}"
            + f" --missing-on-src {src_out}"
            + f" --missing-on-dst {dst_out} --error {error_out}",
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        errors = read(error_out)
        files_stderr = ""
        if errors:
            files_stderr = "/n".join(
                [
                    f"Got {len(errors)} when comparing {source} and {destination}",
                    "Caused by files:",
                    *errors,
                ]
            )
        stderr = (
            result.stderr.decode("utf-8")
            if (not read(src_out) and not read(dst_out) and result.returncode == 1)
            else files_stderr
        )
        os.remove(error_out)
        checks.append(Check(src_out, dst_out, stderr))

    try:
        yield checks
    finally:
        with suppress(FileNotFoundError):
            for check in checks:
                os.remove(check.src_out)
                os.remove(check.dst_out)


def check_paths(paths):
    for pair in paths:
        if len(pair) != 2:
            raise Exception("Source and destination not specified correctly")
        for item in pair:
            split_path = item.split(":")

            # path is local
            if len(split_path) == 1:
                if os.path.expanduser(
                    split_path[0].rstrip(".").rstrip("/")
                ) == os.path.expanduser("~"):
                    raise Exception(f"Local {item} is a root path")
            else:
                if split_path[1].rstrip(".").rstrip("/") == "":
                    raise Exception(f"Remote {item} is a root path")


def sync(source, destination, files_from=None, progress=False):
    args = []
    if files_from is not None:
        args.append(f"--files-from {files_from}")
    if progress:
        args.append("--progress")

    comand = " ".join(["rclone", "sync", source, destination, *args])

    process = subprocess.Popen(comand, shell=True, stdout=subprocess.PIPE)
    if progress:
        pbar = tqdm(total=len(read(files_from)))

        while True:
            line = process.stdout.readline()
            if line:
                line = line.decode("utf-8")
                update = re.findall(
                    r"^Transferred.* ([0,1-9][0-9]*) / [0,1-9][0-9]*", line
                )
                if update:
                    pbar.update(int(update[0]) - pbar.n)
            else:
                pbar.close()
                break
    process.wait()



from rkale.config import get_datasets, global_configuration
from rkale.operations import handle_copy, handle_sync


def handle_pcopy(working_dir, force=False, upstream=False):
    return _apply_to_datasets(handle_copy, working_dir, force=force, upstream=upstream)


def handle_psync(working_dir, force=False, upstream=False):
    return _apply_to_datasets(handle_sync, working_dir, force=force, upstream=upstream)


def _apply_to_datasets(func, working_dir, force=False, upstream=False):
    global_config = global_configuration()
    data_root = global_config["data"]["root"]
    paths = []

    for dataset in get_datasets(working_dir):
        remote = dataset["remote"]
        if "aliases" in global_config and remote in global_config["aliases"]:
            remote = global_config["aliases"][remote]

        paths.append(get_destinations(data_root, remote, dataset["name"], upstream))
    func(paths, force=force)


def get_destinations(data_root, remote, dataset, upstream):
    dest1 = f"{data_root}/{dataset}"
    dest2 = f"{remote}:/{dataset}"
    return (dest1, dest2) if upstream else (dest2, dest1)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rkale']

package_data = \
{'': ['*']}

install_requires = \
['toml>=0.10.2,<0.11.0', 'tqdm>=4.60.0,<5.0.0']

entry_points = \
{'console_scripts': ['rkale = rkale.rkale:main']}

setup_kwargs = {
    'name': 'rkale',
    'version': '0.2.0',
    'description': 'Rclone wrapper to manage multiple datasets in a project',
    'long_description': '# Rkale\n\n## Install\n#### Install Rkale:  \n```bash\npip install rkale\n```\n\n## Configuration\n\n### Global\n\n`~/.config/rkale/rkale.conf`:\n```toml\n[data]\nroot = "path to root data folder"\n\n[aliases]\nwasabi="value to match remote in rclone.conf"\n```\nroot = root folder.  \nIf alases are empty the remote name from the project config is used in the rclone lookup.\n\n### Project\nAdd the rkale tool definition in the pyproject.toml file:  \n\n`~/<project path>/pyproject.toml`:\n```toml\n  [[tool.rkale.dataset]]\n  name = "dataset_1"\n  remote = "remote_1"\n  \n  [[tool.rkale.dataset]]\n  name = "dataset_2"\n  remote = "remote_2"\n  ```\nThe remote specified for the dataset must match a remote in the `rclone.conf` or an alias in the global rkale configuration.\n\n### Project example\n```\n$ rkale psync\n```\nSyncs the datasets specified in the `~/<project path>/pyproject.toml` to be identical with their remotes.  \n```\n$ rkale psync --upstream\n```\nSyncs the remote datasets specified in the `~/<project path>/pyproject.toml` to be identical with their locals.  \n\n### Global example\n```rkale sync <source> <destination>```\nSame as rclone but first checks the result of the operation requires user consent before executing.  \n',
    'author': 'Joar Gruneau',
    'author_email': 'joar@aiwizo.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Aiwizo/rkale',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import os
from functools import lru_cache
from pathlib import Path

import toml


@lru_cache()
def global_configuration():
    return load_configuration(Path.home() / ".config/rkale/rkale.conf")


@lru_cache()
def project_configuration(working_dir):
    config_path = find("pyproject.toml", working_dir)
    config = load_configuration(config_path)
    if "tool" in config and "rkale" in config["tool"]:
        return config["tool"]["rkale"]
    raise Exception(f"No rkale section found in {config_path}")


def get_datasets(working_dir):
    config = project_configuration(working_dir)
    if "dataset" in config:
        return config["dataset"]
    raise Exception("No dataset defined in rkale config")


def get_repository_paths(working_dir):
    data_root = Path(global_configuration()["data"]["root"])
    return {
        dataset["name"]: get_path(data_root, dataset["name"])
        for dataset in get_datasets(working_dir)
    }


def get_path(root, name):
    resolved_path = (root / name).resolve()
    resolved_path.relative_to(root.resolve())
    if resolved_path != root:
        return resolved_path
    raise Exception(f"Path {resolved_path} is outiside data root {root}")


def find(name, path):
    for root, dirs, files in os.walk(path):
        if name in files:
            return os.path.join(root, name)
        raise Exception(f"Could not find file {name} in path {path}")


def load_configuration(path):
    with open(path) as f:
        value = toml.load(f)
        return value
